# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Pandera schemas for psychoanalyze dataframes.

Contains data table schemas of the hierarchical entities described above.
"""
from pandera import (
    Column,
    DataFrameModel,
    DataFrameSchema,
    Index,
    MultiIndex,
    typing,
)

session_dims = ["Monkey", "Date"]
block_stim_dims = ["Amp2", "Width2", "Freq2", "Dur2"]
block_channel_dims = ["Active Channels", "Return Channels"]
block_dims = block_stim_dims + block_channel_dims
point_dims = ["Amp1", "Width1", "Freq1", "Dur1"]

block_index_levels = session_dims + block_dims
points_index_levels = block_index_levels + point_dims


points = DataFrameSchema(
    {
        "n trials": Column(int),
        "Hits": Column(int),
        "Hit Rate": Column(float),
        "logit(Hit Rate)": Column(float),
        "Block": Column(int, required=False),
        "Intensity": Column(float),
    },
)

trials = DataFrameSchema(
    columns={
        "Intensity": Column(float),
        "Result": Column(int),
        "Block": Column(int),
    },
)


blocks = DataFrameSchema(
    columns={"Threshold": Column(dtype=float), "width": Column(dtype=float)},
    index=MultiIndex(
        [
            Index(str, name="Monkey"),
            Index("datetime64", name="Date", coerce=True),
        ]
        + [Index(float, name=dim) for dim in block_stim_dims]
        + [Index(int, name=dim) for dim in block_channel_dims],
    ),
)


psi_animation = DataFrameSchema(
    {
        "Trial": Column(int),
        "Intensity": Column(float),
        "Hit Rate": Column(float),
    },
)


class PsiAnimation(DataFrameModel):
    """Pandera type for psychometric function animation dataset."""

    trial_id: typing.Series[int]
    intensity: typing.Series[float]
    hit_rate: typing.Series[float]


class PsiAnimationFrame(DataFrameModel):
    """Pandera type for a single psychometric function animation frame."""

    intensity: typing.Series[float]
    hit_rate: typing.Series[float]


class Blocks(DataFrameModel):
    """Blocks type for Pandera."""

    slope: float
    threshold: float


class Points(DataFrameModel):
    """Pandera data type."""

    n: int
    Hits: int
    block_id: int


class Trials(DataFrameModel):
    """Trials data type for pandera + mypy type checking."""

    result: int
    intensity: typing.Index[float]



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Bayesian analysis of psychophysical data."""
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def plot(simulated: pd.DataFrame, estimated: pd.Series) -> go.Figure:
    """Plot Psychometric curve to emphasize Bayesian posteriors."""
    combined = pd.concat(
        [simulated.reset_index(), estimated.reset_index()],
        keys=["Simulated", "Estimated"],
        names=["Type"],
    )
    return px.scatter(
        combined.reset_index(),
        x="x",
        y="Hit Rate",
        color="Type",
        template="plotly_white",
    )



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""PsychoAnalyze command line interface."""

import importlib.metadata

import typer
from rich.console import Console

from psychoanalyze.dashboard import app as dash_app

app = typer.Typer()


@app.command()
def main(command: str) -> None:
    """Main commands."""
    if command == "version":
        Console().print(importlib.metadata.version("psychoanalyze"))
    if command == "dash":
        dash_app.app.run(debug=True)



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Psychometric function parameter conversions."""


def intercept_to_location(intercept: float, scale: float) -> float:
    """Convert intercept to location."""
    return -intercept * scale


def location_to_intercept(location: float, scale: float) -> float:
    """Convert location to intercept."""
    return -location / scale


def slope_to_scale(slope: float) -> float:
    """Convert slope to scale."""
    return 1 / slope


def scale_to_slope(scale: float) -> float:
    """Convert scale to slope."""
    return 1 / scale



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Functions for data manipulations at the trial level."""
import json
import random
from pathlib import Path
from typing import TypedDict

import numpy as np
import pandas as pd
from pandera import SeriesSchema, check_output
from sklearn.linear_model import LogisticRegression

from psychoanalyze.data import types

schema = SeriesSchema(bool, name="Test Trials")

data_path = Path("data/trials.csv")

codes = {0: "Miss", 1: "Hit"}


Trial = TypedDict("Trial", {"Result": bool, "Stimulus Magnitude": float})


def generate_trial_index(n_trials: int, options: pd.Index) -> pd.Series:
    """Generate n trials (no outcomes)."""
    return pd.Series(
        [random.choice(options) for _ in range(n_trials)],
        name="Intensity",
    )


@check_output(types.trials)
def generate(
    n_trials: int,
    options: pd.Index,
    params: dict[str, float],
) -> pd.DataFrame:
    """Generate n trials with outcomes."""
    x = generate_trial_index(n_trials, options)
    return pd.DataFrame(
        {
            "Result": [int(random.random() <= psi(x_val, params)) for x_val in x],
            "Intensity": x,
        },
    )


def load(data_path: Path) -> pd.DataFrame:
    """Load trials data from csv."""
    return types.trials.validate(
        pd.read_csv(
            data_path,
            dtype={
                "Result": int,
                "Intensity": float,
                "Block": int,
            },
        ),
    )


def from_store(store_data: str) -> pd.DataFrame:
    """Convert JSON-formatted string to DataFrame."""
    df_dict = json.loads(store_data)
    index_names = df_dict.pop("index_names")
    index = pd.MultiIndex.from_tuples(df_dict["index"])
    trials = pd.DataFrame({"Result": df_dict["data"][0]}, index=index)
    trials.index.names = index_names
    return types.trials.validate(trials)


def to_store(trials: pd.DataFrame) -> str:
    """Convert data to a JSON-formatted string for dcc.Store."""
    data_dict = trials.to_dict(orient="split")
    data_dict["index_names"] = types.points_index_levels
    return json.dumps(data_dict)


def normalize(trials: pd.DataFrame) -> dict[str, pd.DataFrame]:
    """Normalize denormalized trial data."""
    return {
        "Session": trials[["Monkey", "Block"]].drop_duplicates(),
        "Reference Stimulus": trials[["Amp2", "Width2", "Freq2", "Dur2"]],
        "Channel Config": trials[["Active Channels", "Return Channels"]],
        "Test Stimulus": trials[["Amp1", "Width1", "Freq1", "Dur1"]],
    }


def result(p: float) -> bool:
    """Return a trial result given a probability p."""
    return random.random() < p


def results(n: int, p_x: pd.Series) -> list[Trial]:
    """Return a list of trial results in dict format."""
    results = []
    for _ in range(n):
        stimulus_magnitude = random.choice(p_x.index.to_list())
        _result = result(p_x[stimulus_magnitude])
        results.append(
            Trial(
                {
                    "Stimulus Magnitude": stimulus_magnitude,
                    "Result": _result,
                },
            ),
        )
    return results


def labels(results: list[int]) -> list[str]:
    """Convert a list of outcome codes to their labels."""
    return [codes[result] for result in results]


def psi(intensity: float, params: dict[str, float]) -> float:
    """Calculate the value of the psychometric function for a given intensity."""
    gamma = params["Guess Rate"]
    lambda_ = params["Lapse Rate"]
    k = params["Slope"]
    x_0 = params["Threshold"]
    return gamma + (1 - gamma - lambda_) * (1 / (1 + np.exp(-k * (intensity - x_0))))


def moc_sample(n_trials: int, model_params: dict[str, float]) -> pd.DataFrame:
    """Sample results from a method-of-constant-stimuli experiment."""
    x_0 = model_params["x_0"]
    k = model_params["k"]
    intensity_choices = np.linspace(x_0 - 4 / k, x_0 + 4 / k, 7)
    intensities = [float(random.choice(intensity_choices)) for _ in range(n_trials)]
    intensity_index = pd.Index(intensities, name="Intensity")
    results = [
        int(random.random() <= psi(intensity, model_params))
        for intensity in intensities
    ]
    return pd.DataFrame(
        {"Result": pd.Series(results, dtype=int)},
        index=intensity_index,
    )


def fit(trials: pd.DataFrame) -> dict[str, float]:
    """Fit trial data using logistic regression."""
    fits = LogisticRegression().fit(trials[["Intensity"]], trials["Result"])
    return {"Threshold": -fits.intercept_[0], "Slope": fits.coef_[0][0]}



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Constants for dimension labels."""
ref_dims = ["Amp2", "Width2", "Freq2", "Dur2"]
comp_dims = ["Amp1", "Width1", "Freq1", "Dur1"]



"""Layout for Dash dashboard."""

# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

import dash_bootstrap_components as dbc
from dash import dcc, html

from psychoanalyze.dashboard.components import (
    blocks_table,
    link_function,
    model_params,
    points_table,
    simulation_params,
    stimulus_params,
)

font_family = "Comfortaa, Times, serif"
subtitle = "Interactive data simulation & analysis for psychophysics."


input_col = dbc.Col(
    [
        dcc.Upload(
            [
                html.P("Upload CSV/parquet with columns:", className="mb-0"),
                html.P("Block, Intensity, Result."),
                html.P("Your data stays in the browser.", className="mb-0"),
            ],
            id="upload",
            style={
                "borderWidth": "1px",
                "borderStyle": "dashed",
                "borderRadius": "5px",
                "textAlign": "center",
            },
            className="mb-3 p-3",
        ),
        html.H4("Link Function"),
        link_function,
        dbc.Row(
            [
                dbc.Col(html.H4("Model Parameters")),
                dbc.Col(
                    html.I(className="bi bi-info-circle", id="fixed-param-info"),
                    width="auto",
                ),
            ],
            justify="start",
            align="center",
        ),
        dbc.Tooltip(
            "Checked: free parameter\nUnchecked: fixed parameter",
            target="fixed-param-info",
        ),
        model_params,
        html.H4("Stimulus"),
        stimulus_params,
        html.H4("Simulate"),
        dbc.Container(dbc.Row(simulation_params, justify="center")),
    ],
    width=3,
    class_name="mt-4",
    id="input-col",
)

plot_col = dbc.Col(
    [
        dcc.Markdown(
            """
            $$
            \\psi(x) = \\gamma + (1 - \\gamma - \\lambda)F(x)
            $$

            """,
            mathjax=True,
            id="plot-equation",
        ),
        dcc.Graph(id="plot", className="mb-3"),
    ],
    width=5,
)

data_col = dbc.Col(
    [
        html.H4("Blocks", className="mt-3 mb-2"),
        dbc.Container(blocks_table, className="mb-3"),
        dcc.Markdown(
            """
            $$
            F(x) = \\frac{1}{1 + e^{-(\\beta_0 + \\beta_1 x)}}
            $$
            """,
            mathjax=True,
        ),
        html.H4("Points", className="mb-2"),
        dbc.Container(points_table, className="mb-3"),
        dbc.Row(
            [
                dbc.Col(
                    dbc.DropdownMenu(
                        children=[
                            dbc.DropdownMenuItem(
                                "SVG",
                                id={
                                    "type": "img-export",
                                    "name": "svg",
                                },
                            ),
                            dbc.DropdownMenuItem(
                                "PNG",
                                id={
                                    "type": "img-export",
                                    "name": "png",
                                },
                            ),
                            dbc.DropdownMenuItem(
                                "PDF",
                                id={
                                    "type": "img-export",
                                    "name": "pdf",
                                },
                            ),
                        ],
                        label="Download plot",
                        id="figure-download",
                        toggle_style={"border-radius": 5},
                    ),
                    width="auto",
                ),
                dbc.Col(
                    dbc.DropdownMenu(
                        children=[
                            dbc.DropdownMenuItem(
                                "Parquet",
                                id={
                                    "type": "data-export",
                                    "name": "parquet",
                                },
                            ),
                            dbc.DropdownMenuItem(
                                "CSV",
                                id={
                                    "type": "data-export",
                                    "name": "csv",
                                },
                            ),
                            dbc.DropdownMenuItem(
                                "JSON",
                                id={
                                    "type": "data-export",
                                    "name": "json",
                                },
                            ),
                            dbc.DropdownMenuItem(
                                "DuckDB",
                                id={
                                    "type": "data-export",
                                    "name": "duckdb",
                                },
                            ),
                        ],
                        label="Download data",
                        id="data-export",
                        toggle_style={"border-radius": 5},
                    ),
                    width="auto",
                ),
            ],
            justify="around",
        ),
        dcc.Download(id="img-download"),
        dcc.Download(id="data-download"),
    ],
)


layout = dbc.Container(
    [
        dcc.Store(id="points-store"),
        dcc.Store(id="blocks-store"),
        dcc.Store(id="trials-store"),
        dbc.NavbarSimple(
            [
                dbc.NavItem(
                    [
                        dbc.NavLink(
                            "Notebook",
                            href=(
                                """https://nb.psychoanalyze.io/hub/user-redirect/git-pull?repo=https%3A%2F%2Fgithub.com%2Fpsychoanalyze%2Fpsychoanalyze&urlpath=lab%2Ftree%2Fpsychoanalyze%2Fdocs%2Fnotebooks%2Ftutorial.ipynb&branch=main"""
                            ),
                        ),
                        html.I(className="bi bi-journal-code"),
                    ],
                    className="d-flex align-items-center mx-2",
                ),
                dbc.NavItem(
                    [
                        dbc.NavLink(
                            "GitHub",
                            href="https://github.com/psychoanalyze/psychoanalyze",
                        ),
                        html.I(className="bi bi-github"),
                    ],
                    className="d-flex align-items-center mx-2",
                ),
                dbc.NavItem(
                    [
                        dbc.NavLink(
                            "Docs",
                            href="https://docs.psychoanalyze.io",
                        ),
                        html.I(className="bi bi-book"),
                    ],
                    className="d-flex align-items-center mx-2",
                ),
            ],
            brand=dbc.Col(
                [
                    dbc.Row(
                        [
                            dbc.Col(
                                html.Img(
                                    src="assets/logo_transparent_100x100.png",
                                ),
                                width="auto",
                            ),
                            dbc.Col(
                                [
                                    html.H1(
                                        "PsychoAnalyze",
                                        style={"font-family": font_family},
                                    ),
                                    html.P(subtitle),
                                ],
                                align="end",
                            ),
                        ],
                        align="center",
                        justify="start",
                        className="ms-4",
                    ),
                    dbc.Row(),
                ],
            ),
            brand_href="/",
            class_name="mb-2",
            style={"border-radius": "0 0 7px 7px"},
        ),
        dbc.Collapse(
            id="upload-collapse",
        ),
        dbc.Row(
            [
                input_col,
                plot_col,
                data_col,
            ],
        ),
    ],
)



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Global plot settings and generic plot utilities."""

import plotly.express as px
import plotly.graph_objects as go

from psychoanalyze.data import blocks

axis_settings = {
    "ticks": "outside",
    "showgrid": False,
    "showline": True,
    "zeroline": False,
    "title": {"font": {"size": 12, "family": "Arial"}},
}

template = go.layout.Template(
    layout=go.Layout(
        template="plotly_white",
        xaxis=axis_settings,
        yaxis=axis_settings,
        colorway=["#e41a1c", "#377eb8", "#4daf4a"],
        title={"font": {"size": 16, "family": "Arial"}},
        legend={"yanchor": "top", "y": 1, "xanchor": "left", "x": 0.98},
    ),
)

colormap = {"U": "#e41a1c", "Y": "#377eb8", "Z": "#4daf4a"}


labels = {
    "Amp": {
        "x": "Fixed Pulse Width (μs)",
        "y": "Threshold Amplitude (μA)",
    },
    "Width": {
        "x": "Fixed Amplitude (μA)",
        "y": "Threshold Pulse Width (μs)",
    },
}


def standard_logistic() -> go.Scatter:
    """Plot a standard logistic function."""
    return px.line(
        blocks.standard_logistic(),
        y="f(x)",
        template="plotly_white",
        title="$f(x) = \\frac{1}{1 + e^{-x}}$",
    )


def logistic(location: float, scale: float) -> go.Scatter:
    """Plot a logistic function."""
    return px.line(
        blocks.logistic(location, scale),
        y="Ψ(x)",
        template="plotly_white",
    )



"""Main Dash app file.

Contains callbacks.
"""

# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

import base64
import io
import random
import zipfile
from collections.abc import Hashable
from datetime import datetime
from pathlib import Path
from typing import Any

import duckdb
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import pytz
from dash import (
    ALL,
    MATCH,
    Dash,
    Input,
    Output,
    State,
    callback,
    callback_context,
    dcc,
)
from dash_bootstrap_components import icons, themes
from scipy.special import expit, logit

from psychoanalyze.dashboard.layout import layout
from psychoanalyze.data import blocks as pa_blocks
from psychoanalyze.data import points as pa_points

app = Dash(
    __name__,
    external_stylesheets=[
        themes.SUPERHERO,
        icons.BOOTSTRAP,
        "https://fonts.googleapis.com/css2?family=Comfortaa:wght@500&family=Glegoo:wght@700&display=swap",
    ],
)
app.title = "PsychoAnalyze"
app.layout = layout
server = app.server

Records = list[dict[Hashable, Any]]


@callback(
    Output("trials-store", "data"),
    Output({"type": "x-param", "name": "min"}, "value"),
    Output({"type": "x-param", "name": "max"}, "value"),
    Input("upload", "contents"),
    State("upload", "filename"),
    Input({"type": "n-param", "name": ALL}, "value"),
    Input({"type": "param", "name": ALL}, "value"),
)
def update_data(
    contents: str,
    filename: str,
    n_param: list[int],
    param: list[float],
) -> tuple[Records, float, float]:
    """Update points table."""
    n_param_values = [
        "n_levels",
        "n_trials",
        "n_blocks",
    ]
    n_params = pd.Series(n_param, index=n_param_values)
    param = [*param, 0.0, 0.0]
    params = pd.Series(param, index=["x_0", "k", "gamma", "lambda"])
    params["intercept"] = -params["x_0"] / params["k"]
    params["slope"] = 1 / params["k"]
    min_x = (logit(0.01) - params["intercept"]) / params["slope"]
    max_x = (logit(0.99) - params["intercept"]) / params["slope"]
    if callback_context.triggered_id == "upload":
        _, content_string = contents.split(",")
        decoded = base64.b64decode(content_string)
        if "zip" in filename:
            with zipfile.ZipFile(io.BytesIO(decoded)) as z:
                trials = pd.read_csv(z.open("trials.csv"))
        else:
            trials = pd.read_csv(io.StringIO(decoded.decode("utf-8")))
    else:
        points_ix = pd.Index(
            np.linspace(min_x, max_x, n_params["n_levels"]),
            name="Intensity",
        )
        all_trials = {}
        for i in range(n_params["n_blocks"]):
            execution_plan = pd.Index(
                [random.choice(points_ix) for _ in range(n_params["n_trials"])],
                name="Intensity",
            )
            trials_i = pd.Series(
                [
                    int(
                        random.random()
                        < expit(x * params["slope"] + params["intercept"]),
                    )
                    for x in execution_plan
                ],
                name="Result",
                index=execution_plan,
            )
            all_trials[i] = trials_i
        trials = pd.concat(all_trials, names=["Block"]).reset_index()
    return trials.to_dict("records"), min_x, max_x


@callback(
    Output("points-store", "data"),
    Input("trials-store", "data"),
)
def update_points_table(trials: Records) -> Records:
    """Update points table."""
    trials_df = pd.DataFrame.from_records(trials)
    trials_df["Intensity"] = trials_df["Intensity"].astype(float)
    points = pa_points.from_trials(trials_df)
    return points.to_dict("records")


@callback(
    Output("blocks-table", "data"),
    Input("trials-store", "data"),
)
def update_blocks_table(trials: Records) -> Records:
    """Update blocks table."""
    trials_df = pd.DataFrame.from_records(trials)

    blocks = trials_df.groupby("Block").apply(pa_blocks.fit).reset_index()
    blocks["gamma"] = 0.0
    blocks["lambda"] = 0.0
    return blocks.to_dict("records")


@callback(
    Output("points-table", "data"),
    Input("blocks-table", "derived_virtual_selected_rows"),
    Input("points-store", "data"),
)
def filter_points(
    selected_rows: list[int],
    points: Records,
) -> Records:
    """Filter points table."""
    points_df = pd.DataFrame.from_records(points)
    return (
        points_df[points_df["Block"].isin(selected_rows)].to_dict("records")
        if selected_rows
        else points_df.to_dict("records")
    )


@callback(
    Output("plot", "figure"),
    Input({"type": "param", "name": ALL}, "value"),
    Input("points-table", "data"),
    Input("blocks-table", "data"),
    Input("blocks-table", "derived_virtual_selected_rows"),
    State({"type": "x-param", "name": "min"}, "value"),
    State({"type": "x-param", "name": "max"}, "value"),
)
def update_fig(  # noqa: PLR0913
    param: list[float],
    points: Records,
    blocks: Records,
    selected_rows: list[int],
    min_x: float,
    max_x: float,
) -> go.Figure:
    """Update plot and tables based on data store and selected view."""
    param = [*param, 0.0, 0.0]
    params = pd.Series(param, index=["x_0", "k", "gamma", "lambda"])
    params["intercept"] = -params["x_0"] / params["k"]
    params["slope"] = 1 / params["k"]
    blocks = [blocks[i] for i in selected_rows] + [
        {"Block": "Model"} | params.to_dict(),
    ]
    x = pd.Index(
        np.linspace(min_x, max_x, 100),
        name="Intensity",
    )
    fits = pd.concat(
        {
            block["Block"]: pd.Series(
                expit(x.to_numpy() * block["slope"] + block["intercept"]),
                name="Hit Rate",
                index=x,
            )
            for block in blocks
        },
        names=["Block"],
    ).reset_index()
    fits["Block"] = fits["Block"].astype(str)
    points_df = pd.DataFrame.from_records(points)
    points_df["Block"] = points_df["Block"].astype(str)
    fits_fig = px.line(
        fits,
        x="Intensity",
        y="Hit Rate",
        color="Block",
    )
    results_fig = px.scatter(
        points_df,
        x="Intensity",
        y="Hit Rate",
        size="n trials",
        color="Block",
        template="plotly_white",
    )
    return results_fig.add_traces(fits_fig.data)


@callback(
    Output("img-download", "data"),
    Input({"type": "img-export", "name": ALL}, "n_clicks"),
    State("plot", "figure"),
    prevent_initial_call=True,
)
def export_image(
    export_clicked: int,  # noqa: ARG001
    fig: go.Figure,
) -> dict[str, str | bool | bytes]:
    """Export image."""
    format_suffix = callback_context.triggered_id["name"]
    return {
        "base64": True,
        "content": base64.b64encode(
            go.Figure(fig)
            .update_layout(showlegend=False)
            .to_image(format=format_suffix, width=500, height=500),
        ).decode("utf-8"),
        "filename": f"fig.{format_suffix}",
    }


@callback(
    Output("data-download", "data"),
    Input({"type": "data-export", "name": ALL}, "n_clicks"),
    State("points-table", "data"),
    State("blocks-table", "data"),
    State("trials-store", "data"),
    prevent_initial_call=True,
)
def export_data(
    export_clicked: int,  # noqa: ARG001
    points: Records,
    blocks: Records,
    trials: Records,
) -> dict[str, Any | None]:
    """Export image."""
    format_suffix = callback_context.triggered_id["name"]
    points_df = pd.DataFrame.from_records(points)
    blocks_df = pd.DataFrame.from_records(blocks)
    trials_df = pd.DataFrame.from_records(trials)
    if format_suffix == "csv":
        zip_buffer = io.BytesIO()

        with zipfile.ZipFile(
            zip_buffer,
            mode="a",
            compression=zipfile.ZIP_DEFLATED,
            allowZip64=False,
        ) as zip_file:

            def write_file(df: pd.DataFrame, name: str) -> None:
                buffer = io.StringIO()
                df.to_csv(buffer, index=False)
                zip_file.writestr(name, buffer.getvalue())

            data = {
                "points": points_df,
                "blocks": blocks_df,
                "trials": trials_df,
            }
            for level, points_df in data.items():
                write_file(points_df, f"{level}.csv")

        zip_buffer.seek(0)
        zip_bytes = zip_buffer.read()

        base64_bytes = base64.b64encode(zip_bytes)
        base64_string = base64_bytes.decode("utf-8")
        timestamp = datetime.now(tz=pytz.timezone("America/Chicago")).strftime(
            "%Y-%m-%d_%H%M",
        )
        download = {
            "base64": True,
            "content": base64_string,
            "filename": f"{timestamp}_psychoanalyze.zip",
        }

    elif format_suffix == "json":
        download = dcc.send_data_frame(points_df.to_json, "data.json")
    elif format_suffix == "parquet":
        download = dcc.send_data_frame(points_df.to_parquet, "data.parquet")
    elif format_suffix == "duckdb":
        connection = duckdb.connect("psychoanalyze.duckdb")
        connection.sql("CREATE TABLE points AS SELECT * FROM _data")
        connection.close()
        with Path("psychoanalyze.duckdb").open("rb") as f:
            download = {
                "base64": True,
                "content": base64.b64encode(f.read()).decode("utf-8"),
                "filename": f"psychoanalyze.{format_suffix}",
            }

    return download


@callback(
    Output("F-eqn", "is_open"),
    Output("show-eqn", "children"),
    Output("plot-equation", "children"),
    Input("show-eqn", "n_clicks"),
)
def toggle_eqn(n_clicks: int) -> tuple[bool, str, str]:
    """Toggle equation."""
    equation_abstracted = """
    $$
    \\psi(x) = \\gamma + (1 - \\gamma - \\lambda)F(x)
    $$
    """
    if n_clicks:
        n_clicks_is_even = n_clicks % 2 == 0
        label = "Show F(x) ▾ " if n_clicks_is_even else "Hide F(x) ▴ "
        equation_expanded = """
        $$
        \\psi(x) = \\frac{\\gamma + (1 - \\gamma - \\lambda)}{1 + e^{-k(x - x_0)}}
        $$
        """
        equation = equation_abstracted if n_clicks_is_even else equation_expanded
        return not n_clicks_is_even, label, equation
    return False, "Show F(x) ▾ ", equation_abstracted


@callback(
    Output({"type": "param", "name": MATCH}, "disabled"),
    Input({"type": "free", "name": MATCH}, "value"),
    prevent_initial_call=True,
)
def toggle_param(free: bool) -> bool:  # noqa: FBT001
    """Toggle parameter."""
    return not free


@callback(
    Output({"type": "param", "name": ALL}, "value"),
    Input("preset", "value"),
    State({"type": "param", "name": ALL}, "value"),
    prevent_initial_call=True,
)
def set_params_to_preset(
    preset: str,
    param: list[float],
) -> list[float]:
    """Set parameters to preset values."""
    presets = {
        "standard": [0.0, 1.0, 0.0, 0.0],
        "non-standard": [10, 2, 0.2, 0.1],
        "2AFC": [0, 1, 0.5, 0.0],
    }
    return presets.get(preset, param)


@callback(
    Output({"type": "free", "name": ALL}, "value"),
    Input("preset", "value"),
    State({"type": "free", "name": ALL}, "value"),
    prevent_initial_call=True,
)
def set_fixed_params_to_preset(
    preset: str,
    param: list[bool],
) -> list[bool]:
    """Set parameters to preset values."""
    presets = {
        "standard": [True] * 4,
        "non-standard": [True] * 4,
        "2AFC": [True, True, False, True],
    }
    return presets.get(preset, param)


@callback(
    Output("points-table", "page_size"),
    Input({"type": "n-param", "name": "n-levels"}, "value"),
)
def update_page_size(n_levels: int) -> int:
    """Update page size."""
    return n_levels


if __name__ == "__main__":
    app.run(debug=True)



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Sigmoid functions used in the psychometric function."""
from typing import Any

import numpy as np


def weibull(
    x: np.ndarray[Any, np.dtype[np.floating[Any]]],
    alpha: float,
    beta: float,
) -> float:
    """Calculate psi using Weibull function."""
    return 1 - np.exp(-((x / alpha) ** beta))


def gumbel(x: np.ndarray[Any, np.dtype[Any]], alpha: float, beta: float) -> float:
    """Calculate psi using gumbel function."""
    return 1 - np.exp(-(10 ** (beta * (x - alpha))))


def quick(x: float, alpha: float, beta: float) -> float:
    """Calculate psi using quick function."""
    return 1 - 2 ** (-((x / alpha) ** beta))


def log_quick(x: float, alpha: float, beta: float) -> float:
    """Calculate psi using log_quick function."""
    return 1 - 2 ** (-(10 ** (beta * (x - alpha))))



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Utilities for session-level data.

**Sessions** represent a single day of experiments performed by a subject. It may
contain several blocks.
"""
from pathlib import Path

import pandas as pd

from psychoanalyze.data import blocks

dims = ["Monkey", "Date"]
index_levels = dims


def generate(n: int) -> list[int]:
    """Generate session-level data."""
    return list(range(n))


def cache_results(sessions: pd.DataFrame) -> None:
    """Save session data to csv."""
    sessions.to_csv("data/normalized/sessions.csv", index=False)


def from_trials_csv(path: Path) -> pd.DataFrame:
    """Aggregate to session level from trial-level data."""
    return pd.read_csv(path)[["Monkey", "Date"]].drop_duplicates()


def day_marks(subjects: pd.DataFrame, sessions: pd.DataFrame, monkey: str) -> dict:
    """Calculate days since surgery date for a given subject."""
    surgery_date = pd.to_datetime(
        subjects.loc[subjects["Monkey"] == monkey, "Surgery Date"],
    )[0]
    sessions = sessions[sessions["Monkey"] == "U"]
    sessions["Days"] = (pd.to_datetime(sessions["Date"]) - surgery_date).dt.days
    return {sessions.loc[i, "Days"]: sessions.loc[i, "Date"] for i in sessions.index}


def days(sessions: pd.DataFrame, subjects: pd.DataFrame) -> pd.Series:
    """Calculate days since surgery date."""
    sessions_subjects = sessions.join(subjects, on="Monkey")
    return (
        pd.to_datetime(sessions_subjects.index.get_level_values("Date"))
        - sessions_subjects["Surgery Date"]
    ).dt.days


def n_trials(trials: pd.DataFrame) -> pd.DataFrame:
    """Count trials per session."""
    return trials.groupby(["Monkey", "Date"])[["Result"]].count()


def load(data_dir: Path) -> pd.DataFrame:
    """Load session-level data from csv."""
    return pd.read_csv(data_dir / "sessions.csv", index_col=["Monkey", "Date"])


def generate_trials(
    n_trials: int,
    model_params: dict[str, float],
    n_days: int,
) -> pd.DataFrame:
    """Generate trial-level data for session-level context."""
    return pd.concat(
        {day: blocks.generate_trials(n_trials, model_params) for day in range(n_days)},
        names=["Block"],
    )



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""The PsychoAnalyze dashboard is powered by Plotly/Dash.

- `app.py` is the main file which contains the infrastructure for the Dash app.
and callback functions in their entirety.

- `layout.py` specifies the layout of the HTML and Dash components that comprise
the app.

- `components.py` contains more complex and/or reusable components used in the app.

"""



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Empirical Distribution Functions (eCDF)."""
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def plot(blocks: pd.DataFrame, param: str) -> go.Figure:
    """Plot empirical cumulative distrubtion function (eCDF) of fitted params."""
    return px.ecdf(
        blocks.reset_index(),
        x=param,
        color=blocks.get("Monkey"),
    ).update_layout(xaxis_title=param)



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Utilities for points-level data.

**Points** correspond to the aggregate measures of method-of-constant-stimuli
experiments at each stimulus level measured. For example, a block that samples 8
stimulus intensity levels would have 8 corresponding points.
"""
from pathlib import Path

import cmdstanpy as stan
import numpy as np
import pandas as pd
import plotly.express as px
from dash import dash_table
from pandera import check_io, check_output
from plotly import graph_objects as go
from scipy.special import expit, logit
from scipy.stats import logistic

from psychoanalyze.data import trials as pa_trials
from psychoanalyze.data import types

index_levels = ["Amp1", "Width1", "Freq1", "Dur1"]


@check_io(trials=types.trials, out=types.points)
def from_trials(trials: pd.DataFrame) -> pd.DataFrame:
    """Aggregate point-level measures from trial data."""
    points = trials.groupby(["Block", "Intensity"])["Result"].agg(["count", "sum"])
    points = points.rename(columns={"count": "n trials", "sum": "Hits"})
    points["Hit Rate"] = points["Hits"] / points["n trials"]
    points["logit(Hit Rate)"] = logit(points["Hit Rate"])
    return points.reset_index()


@check_output(types.points)
def load(data_path: Path) -> pd.DataFrame:
    """Load points data from csv."""
    trials = pa_trials.load(data_path)
    return from_trials(trials)


def prep_fit(points: pd.DataFrame, dimension: str = "Amp1") -> dict:
    """Transform points data for numpy-related fitting procedures."""
    points = points.reset_index()
    return {
        "X": len(points),
        "x": points[f"{dimension}"].to_numpy(),
        "N": points["n"].to_numpy(),
        "hits": points["Hits"].to_numpy(),
    }


def model() -> stan.CmdStanModel:
    """Instantiate Stan binomial regression model."""
    return stan.CmdStanModel(stan_file="models/binomial_regression.stan")


def hits(
    n: pd.Series,
    params: dict[str, float],
) -> pd.Series:
    """Sample list of n hits from a list of intensity values."""
    p = logistic.cdf(n.index.to_numpy(), params["Threshold"], params["Slope"])
    psi = params["Guess Rate"] + (1.0 - params["Guess Rate"] - params["Lapse Rate"]) * p
    return pd.Series(
        np.random.default_rng().binomial(
            n,
            psi,
            len(n),
        ),
        index=n.index,
        name="Hits",
    )


def generate(
    n_trials: int,
    options: pd.Index,
    params: dict[str, float],
) -> pd.DataFrame:
    """Generate points-level data."""
    n = generate_n(n_trials, options)
    _hits = hits(
        n,
        params,
    )
    points = pd.concat([n, _hits], axis=1)
    _hit_rate = hit_rate(points)
    logit_hit_rate = pd.Series(
        logit(_hit_rate),
        name="logit(Hit Rate)",
        index=n.index,
    )
    return pd.concat([points, _hit_rate, logit_hit_rate], axis=1)


def generate_point(n: int, p: float) -> int:
    """Sample n hits from n trials and probability p from binomial dist."""
    return np.random.default_rng().binomial(n, p)


def datatable(data: pd.DataFrame) -> dash_table.DataTable:
    """Convert dataframe to Dash DataTable-friendly format."""
    return dash_table.DataTable(
        data.reset_index()[["Amp1", "Hit Rate", "n"]].to_dict("records"),
        columns=[
            {
                "id": "Amp1",
                "name": "Amp1",
                "type": "numeric",
                "format": dash_table.Format.Format(
                    precision=2,
                    scheme=dash_table.Format.Scheme.fixed,
                ),
            },
            {
                "id": "Hit Rate",
                "name": "Hit Rate",
                "type": "numeric",
                "format": dash_table.Format.Format(
                    precision=2,
                    scheme=dash_table.Format.Scheme.fixed,
                ),
            },
            {
                "id": "n",
                "name": "n",
                "type": "numeric",
            },
        ],
        id="experiment-psych-table",
    )


def from_store(store_data: str) -> pd.DataFrame:
    """Get points-level measures from trials-level data store."""
    trials = pa_trials.from_store(store_data)
    return from_trials(trials)


def combine_plots(fig1: go.Figure, fig2: go.Figure) -> go.Figure:
    """Combine two points-level plots. Possible duplicate."""
    return go.Figure(data=fig1.data + fig2.data)


def n(trials: pd.Series) -> pd.Series:
    """Count trials at each point."""
    return pd.Series(trials.value_counts(), name="n")


def generate_n(n_trials: int, options: pd.Index) -> pd.Series:
    """Simulate how many trials were performed per intensity level."""
    return n(pa_trials.generate_trial_index(n_trials, options))


def to_block(points: pd.DataFrame) -> pd.DataFrame:
    """Aggregate to block-level measures from points-level data."""
    return points.groupby(level="Block").sum()


def psi(
    x: pd.Index,
    params: dict[str, float],
) -> pd.Series:
    """Calculate psi for an array of intensity levels x."""
    return pd.Series(
        params["gamma"]
        + (1 - params["gamma"] - params["lambda"])
        * expit(params["x_0"] + params["k"] * x),
        index=x,
        name="p(x)",
    )


def plot(points: pd.DataFrame, y: str) -> go.Figure:
    """Plot the psychometric function."""
    return px.scatter(
        points.reset_index(),
        x="Intensity",
        y=y,
        size="n",
        color="Block",
        template="plotly_white",
    )


def hit_rate(df: pd.DataFrame) -> pd.Series:
    """Calculate hit rate from hits and number of trials."""
    return pd.Series(df["Hits"] / df["n"], name="Hit Rate")


def transform(hit_rate: float, y: str) -> float:
    """Logit transform hit rate."""
    return logit(hit_rate) if y == "alpha" else hit_rate


def generate_index(n_levels: int, x_range: list[float]) -> pd.Index:
    """Generate evenly-spaced values along the modulated stimulus dimension."""
    min_x = x_range[0]
    max_x = x_range[1]
    return pd.Index(np.linspace(min_x, max_x, n_levels), name="Intensity")



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Data transformation functions for subject-level data."""
import string
from pathlib import Path

import pandas as pd

from psychoanalyze.data import sessions


def load(data_path: Path) -> pd.DataFrame:
    """Load subject data from csv."""
    return pd.read_csv(
        data_path / "subjects.csv",
        index_col="Monkey",
        parse_dates=["Surgery Date"],
    )


def generate_letter_names(n_subjects: int) -> list[str]:
    """Generate a list of dummy subjects using capital letters in alph. order."""
    return list("ABCDEFG"[:n_subjects])


def generate_trials(
    n_trials: int,
    model_params: dict[str, float],
    n_days: int,
    n_subjects: int,
) -> pd.DataFrame:
    """Generate trial-level data, including subject-level info."""
    return pd.concat(
        {
            subj: sessions.generate_trials(n_trials, model_params, n_days)
            for subj in string.ascii_uppercase[:n_subjects]
        },
        names=["Subject"],
    )



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Strength-duration analysis.

Contains functions assessing the relationship
between the amplitude and the time course of the stimulus.
"""
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from psychoanalyze.plot import labels, template


def from_blocks(blocks: pd.DataFrame, dim: str) -> pd.DataFrame:
    """Calculate strength-duration measures from block data."""
    if dim == "Amp":
        ylabel = "Threshold Amplitude (μA)"
        xlabel = "Fixed Pulse Width (μs)"
    elif dim == "Width":
        ylabel = "Fixed Amplitude (μA)"
        xlabel = "Threshold Pulse Width (μs)"

    blocks[ylabel] = blocks["Threshold"]
    blocks[xlabel] = blocks["Fixed Magnitude"]
    return blocks.drop(columns=["Threshold", "Fixed Magnitude"])


def plot(
    blocks: pd.DataFrame,
    dim: str,
    x_data: list[float],
    y_data: list[float],
) -> go.Figure:
    """Plot strength-duration curve given detection data."""

    def _get_labels_given_dim(
        labels: dict[str, dict[str, str]],
        dim: str,
    ) -> dict[str, str]:
        """Get appropriate axis labels for different choices of modulated dimension."""
        return {"x": labels[dim]["x"], "y": labels[dim]["y"]}

    labels_given_dim = _get_labels_given_dim(labels=labels, dim=dim)
    x = labels_given_dim["x"]
    y = labels_given_dim["y"]
    if blocks is not None:
        sd_df = blocks[blocks["Dimension"] == dim]
    else:
        sd_df = pd.DataFrame({x: x_data, y: y_data})
    return px.scatter(
        sd_df,
        x=x,
        y=y,
        template=template,
    )



"""Components for Dash dashboard."""

# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
#  PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

import dash_bootstrap_components as dbc
from dash import dash_table, dcc

model_params = dbc.Container(
    [
        dbc.Row(
            [
                dbc.Col(
                    dcc.Markdown("$x_0$", mathjax=True),
                    width=2,
                    style={"text-align": "center"},
                    align="center",
                ),
                dbc.Col(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(
                                dbc.Checkbox(
                                    id={"type": "free", "name": "x_0"},
                                    value=True,
                                ),
                            ),
                            dbc.Input(
                                id={"type": "param", "name": "x_0"},
                                type="number",
                                value=0.0,
                            ),
                        ],
                        size="sm",
                        class_name="p-0",
                    ),
                    width=4,
                    style={"text-align": "center"},
                    align="center",
                ),
                dbc.Col(
                    dcc.Markdown("$k$", mathjax=True),
                    width=2,
                    align="center",
                    style={"text-align": "center"},
                ),
                dbc.Col(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(
                                dbc.Checkbox(
                                    value=True,
                                    id={"type": "free", "name": "k"},
                                ),
                            ),
                            dbc.Input(
                                id={"type": "param", "name": "k"},
                                type="number",
                                value=1.0,
                                step=0.1,
                            ),
                        ],
                        size="sm",
                        class_name="p-0",
                    ),
                    width=4,
                    align="center",
                    style={"text-align": "center"},
                ),
            ],
            class_name="g-0 mb-0",
        ),
        dbc.Row(
            [
                dbc.Col(
                    dcc.Dropdown(
                        id="preset",
                        placeholder="Presets",
                        options=[
                            {
                                "label": "Standard",
                                "value": "standard",
                            },
                            {
                                "label": "Non Standard",
                                "value": "non-standard",
                            },
                        ],
                    ),
                    width=7,
                    class_name="mt-2",
                ),
            ],
            justify="center",
        ),
    ],
    className="mb-2",
)

link_function = dbc.Container(
    [
        dbc.Row(
            [
                dbc.Col(
                    dcc.Dropdown(
                        id="f",
                        options=[{"label": "Logistic", "value": "expit"}],
                        value="expit",
                        className="mb-1",
                    ),
                ),
                dbc.Col(
                    dbc.Button(
                        style={"border-radius": 3},
                        id="show-eqn",
                        class_name="btn btn-outline-info",
                        outline=True,
                    ),
                    width=6,
                ),
            ],
        ),
        dbc.Collapse(
            dcc.Markdown(
                """
                    $$
                    F(x) = \\frac{1}{1 + e^{-k(x-x_0)}}
                    $$""",
                mathjax=True,
            ),
            is_open=False,
            id="F-eqn",
        ),
    ],
    className="mb-2",
)

stimulus_params = dbc.Container(
    [
        dbc.Row(
            [
                dbc.Col(
                    [
                        dbc.FormText("min"),
                        dbc.Input(
                            id={"type": "x-param", "name": "min"},
                            type="number",
                            disabled=True,
                        ),
                    ],
                ),
                dbc.Col(
                    [
                        dbc.FormText("n levels"),
                        dbc.Input(
                            id={"type": "n-param", "name": "n-levels"},
                            type="number",
                            value=7,
                        ),
                    ],
                ),
                dbc.Col(
                    [
                        dbc.FormText("max"),
                        dbc.Input(
                            id={"type": "x-param", "name": "max"},
                            type="number",
                            disabled=True,
                        ),
                    ],
                ),
            ],
            className="mb-1 g-0",
        ),
    ],
    class_name="mb-3",
)


simulation_params = dbc.Col(
    [
        dbc.InputGroup(
            [
                dbc.Input(
                    id={"type": "n-param", "name": "n-trials"},
                    type="number",
                    value=100,
                ),
                dbc.InputGroupText("trials per block"),
            ],
        ),
        dbc.InputGroup(
            [
                dbc.Input(
                    id={"type": "n-param", "name": "n-blocks"},
                    type="number",
                    value=5,
                ),
                dbc.InputGroupText("blocks"),
            ],
        ),
    ],
    width="9",
)

points_table = dash_table.DataTable(
    id="points-table",
    columns=[
        {
            "name": "Block",
            "id": "Block",
            "type": "numeric",
        },
        {
            "name": "Intensity",
            "id": "Intensity",
            "type": "numeric",
            "format": dash_table.Format.Format(
                precision=2,
                scheme=dash_table.Format.Scheme.fixed,
            ),
        },
        {
            "name": "n trials",
            "id": "n trials",
            "type": "numeric",
        },
        {
            "name": "p(x)",
            "id": "p(x)",
            "type": "numeric",
            "format": dash_table.Format.Format(
                precision=2,
                scheme=dash_table.Format.Scheme.fixed,
            ),
        },
        {
            "name": "Hits",
            "id": "Hits",
            "type": "numeric",
        },
        {
            "name": "Hit Rate",
            "id": "Hit Rate",
            "type": "numeric",
            "format": dash_table.Format.Format(
                precision=2,
                scheme=dash_table.Format.Scheme.fixed,
            ),
        },
    ],
    style_data={"color": "black"},
    style_header={"color": "black"},
    page_action="native",
)

blocks_table = dash_table.DataTable(
    id="blocks-table",
    row_selectable="multi",
    selected_rows=[0, 2],
    columns=[
        {
            "name": "Block",
            "id": "Block",
            "type": "numeric",
        },
        {
            "name": "intercept β₀",
            "id": "intercept",
            "type": "numeric",
            "format": dash_table.Format.Format(
                precision=2,
                scheme=dash_table.Format.Scheme.fixed,
            ),
        },
        {
            "name": "slope β₁",
            "id": "slope",
            "type": "numeric",
            "format": dash_table.Format.Format(
                precision=2,
                scheme=dash_table.Format.Scheme.fixed,
            ),
        },
    ],
    style_data={"color": "black"},
    style_header={"color": "black"},
)



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Top-level package for `psychoanalyze`.

Modules:

- [`plot`][psychoanalyze.plot]: Global plot settings and generic plot utilities.
- [`sigmoids`][psychoanalyze.sigmoids]: Implementations of psychometric sigmoid
functions.

Subpackages:

- [`data`][psychoanalyze.data]: Submodules in [`psychoanalyze.data`][psychoanalyze.data]
contain data manipulation and transformation functions that relate to a level in the
data hierarchy.
- [`analysis`][psychoanalyze.analysis]: Submodules in
[`psychoanalyze.analysis`][psychoanalyze.analysis] contain data manipulation and
transformation functions, often corresponding to a level in the data hierarchy.
"""



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Data modules and general-purpose data transformation utilities.

Submodules:

- [`psychoanalyze.data.blocks`][psychoanalyze.data.blocks]
- [`psychoanalyze.data.points`][psychoanalyze.data.points]
- [`psychoanalyze.data.trials`][psychoanalyze.data.trials]
- [`psychoanalyze.data.sessions`][psychoanalyze.data.sessions]
- [`psychoanalyze.data.subjects`][psychoanalyze.data.subjects]
- [`psychoanalyze.data.types`][psychoanalyze.data.types]
"""



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Block-level data utilities.

**Blocks** are the most analytically significant objects in the PsychoAnalyze
data hierarchy. They represent a specific set of experimental conditions and generally
correspond to a single fit of the psychometric function.
"""
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from scipy.special import expit
from scipy.stats import logistic as scipy_logistic
from sklearn.linear_model import LogisticRegression

from psychoanalyze.data import (
    sessions,
    stimulus,
    subjects,
    trials,
)
from psychoanalyze.plot import template

dims = ["Amp2", "Width2", "Freq2", "Dur2", "Active Channels", "Return Channels"]
index_levels = dims


def generate(
    n_trials_per_level: int,
    x_min: float,
    x_max: float,
    n_levels: int,
) -> pd.DataFrame:
    """Generate block-level data."""
    index = pd.Index(np.linspace(x_min, x_max, n_levels), name="x")
    n = [n_trials_per_level] * len(index)
    p = scipy_logistic.cdf(index)
    return pd.DataFrame(
        {"n": n, "Hits": np.random.default_rng().binomial(n, p)},
        index=index,
    )


def plot_fits(blocks: pd.DataFrame) -> go.Figure:
    """Plot fits."""
    x = np.linspace(-3, 3, 100)
    y = expit(x)
    return px.line(blocks.reset_index(), x=x, y=y)


def load(data_path: Path) -> pd.DataFrame:
    """Load block data from csv."""
    full_path = data_path / "blocks.csv"
    channel_config = ["Active Channels", "Return Channels"]
    blocks = pd.read_csv(full_path / "blocks.csv", parse_dates=["Date"]).set_index(
        sessions.dims + stimulus.ref_dims + channel_config,
    )
    blocks["Block"] = days(blocks, subjects.load(full_path))
    return blocks


def days(blocks: pd.DataFrame, intervention_dates: pd.DataFrame) -> pd.Series:
    """Calculate days for block-level data. Possible duplicate."""
    blocks = blocks.join(intervention_dates, on="Subject")
    days = pd.Series(
        blocks.index.get_level_values("Date") - blocks["Surgery Date"],
    ).dt.days
    days.name = "Days"
    return days


def n_trials(trials: pd.DataFrame) -> pd.Series:
    """Calculate n trials for each block."""
    session_cols = ["Subject", "Date"]
    ref_stim_cols = ["Amp2", "Width2", "Freq2", "Dur2"]
    channel_config = ["Active Channels", "Return Channels"]
    return trials.groupby(session_cols + ref_stim_cols + channel_config)[
        "Result"
    ].count()


def is_valid(block: pd.DataFrame) -> bool:
    """Determine if curve data is valid."""
    return any(block["Hit Rate"] > 0.5) & any(block["Hit Rate"] < 0.5)  # noqa: PLR2004


def subject_counts(data: pd.DataFrame) -> pd.DataFrame:
    """Determine how many subjects are in the data."""
    summary = (
        data.index.get_level_values("Subject").value_counts().rename("Total Blocks")
    )
    summary.index.name = "Subject"
    return summary


def fit(trials: pd.DataFrame) -> pd.Series:
    """Fit logistic regression to trial data."""
    fit = LogisticRegression().fit(
        trials[["Intensity"]],
        trials["Result"],
    )
    intercept = fit.intercept_[0]
    slope = fit.coef_[0][0]
    return pd.Series({"intercept": intercept, "slope": slope})


def generate_trials(n_trials: int, model_params: dict[str, float]) -> pd.DataFrame:
    """Generate trials for block-level context."""
    return trials.moc_sample(n_trials, model_params)


def from_points(points: pd.DataFrame) -> pd.DataFrame:
    """Aggregate block measures from points data."""
    return points.groupby("BlockID")[["n"]].sum()


def plot_thresholds(blocks: pd.DataFrame) -> go.Figure:
    """Plot longitudinal threshold data.

    Args:
        blocks: Block-level DataFrame.

    Returns:
        A plotly Graph Object.
    """
    return px.scatter(
        transform_errors(blocks),
        x="Block",
        y="50%",
        error_y="err+",
        error_y_minus="err-",
        color="Subject",
        color_discrete_map={"U": "#e41a1c", "Y": "#377eb8", "Z": "#4daf4a"},
        template=template,
    )


def transform_errors(fit: pd.DataFrame) -> pd.DataFrame:
    """Transform errors from absolute to relative."""
    fit["err+"] = fit["95%"] - fit["50%"]
    fit["err-"] = fit["50%"] - fit["5%"]
    return fit.drop(columns=["95%", "5%"])


def reshape_fit_results(fits: pd.DataFrame, x: pd.Index, y: str) -> pd.DataFrame:
    """Reshape fit params for plotting."""
    rows = [f"{y}[{i}]" for i in range(1, len(x) + 1)]
    param_fits = fits.loc[
        rows,  # row eg 'p[1]:p[8]'
        ["5%", "50%", "95%"],  # col
    ]
    param_fits = transform_errors(param_fits)
    param_fits = param_fits.rename(columns={"50%": y})
    param_fits.index = x
    return param_fits


def standard_logistic() -> pd.Series:
    """Generate points for a line trace of a standard logistic function."""
    x = pd.Index(np.linspace(-3, 3, 100), name="x")
    y = expit(x)
    return pd.Series(y, index=x, name="f(x)")


def logistic(location: float, scale: float) -> pd.Series:
    """Generate points for a line trace of a logistic function."""
    x_min = (location - 4) * scale
    x_max = (location + 4) * scale
    x = pd.Index(np.linspace(x_min, x_max, 100), name="Intensity")
    y = expit((x - location) / scale)
    return pd.Series(y, index=x, name="Ψ(x)")



# Copyright 2023 Tyler Schlichenmeyer

# This file is part of PsychoAnalyze.
# PsychoAnalyze is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.

# PsychoAnalyze is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# PsychoAnalyze. If not, see <https://www.gnu.org/licenses/>.

"""Test functions related to Weber's Law analysis.

Contains functions assessing how discriminability of two stimuli
relates to the baseline intensities of the stimuli according to Weber's Law.
"""
from pathlib import Path

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def plot(
    data: pd.DataFrame,
    trendline: str = "ols",
    error_y: str | None = None,
) -> go.Figure:
    """Plot data according to Weber's Law."""
    _trendline = "ols" if trendline else None
    return px.scatter(
        data.reset_index(),
        x="Reference Charge (nC)",
        y="Difference Threshold (nC)",
        error_y="err+" if error_y == "error bars" else None,
        error_y_minus="err-" if error_y == "error bars" else None,
        color="Monkey",
        symbol="Dimension",
        trendline=_trendline,
        template="plotly_white",
        hover_data=["Date"],
    )


def aggregate(data: pd.DataFrame) -> pd.DataFrame:
    """Calculate agg stats for Weber data."""
    return (
        data.groupby(["Monkey", "Dimension", "Reference Charge (nC)"])[
            "Difference Threshold (nC)"
        ]
        .agg(["mean", "count", "std"])
        .rename(columns={"mean": "Difference Threshold (nC)"})
    )


def load(path: Path) -> pd.DataFrame:
    """Load weber file from a csv."""
    weber = pd.read_csv(path, parse_dates=["Date"])
    weber["err+"] = (
        weber["location_CI_5"] * weber["Fixed_Param_Value"] / 1000
    ) - weber["Threshold_Charge_nC"]
    weber["err-"] = (
        weber["Threshold_Charge_nC"]
        - (weber["location_CI_95"]) * weber["Fixed_Param_Value"] / 1000
    )
    return weber
